@@ -1,7 +1,6 @@
 /******************************************************************************
- * $Id$
  *
- * Project:  MapServer
+ * Project:  MapCache
  * Author:   Thomas Bonfort and the MapServer team.
  *
  ******************************************************************************
@@ -46,34 +45,11 @@
 
 #include "errors.h"
 
-#if 0
-#ifdef USE_GDAL
-#include <gdal.h>
-#include <cpl_conv.h>
-#endif
-#endif
 
 #include <assert.h>
 #include <time.h>
 #include <apr_time.h>
 
-#ifdef USE_PCRE
-#include <pcre.h>
-#else
-#include <regex.h>
-#endif
-
-#ifdef USE_MEMCACHE
-#include <apr_memcache.h>
-#endif
-
-#ifdef USE_COUCHBASE
-#include <libcouchbase/couchbase.h>
-#endif
-
-#ifdef USE_RIAK
-#include <riack.h>
-#endif
 
 #define MAPCACHE_SUCCESS 0
 #define MAPCACHE_FAILURE 1
@@ -113,44 +89,16 @@ typedef struct mapcache_feature_info mapcache_feature_info;
 typedef struct mapcache_request_get_feature_info mapcache_request_get_feature_info;
 typedef struct mapcache_map mapcache_map;
 typedef struct mapcache_http_response mapcache_http_response;
-typedef struct mapcache_source_wms mapcache_source_wms;
-#if 0
-typedef struct mapcache_source_gdal mapcache_source_gdal;
-#endif
-typedef struct mapcache_cache_disk mapcache_cache_disk;
-typedef struct mapcache_cache_composite mapcache_cache_composite;
-typedef struct mapcache_cache_fallback mapcache_cache_fallback;
-typedef struct mapcache_cache_multitier mapcache_cache_multitier;
-typedef struct mapcache_cache_rest mapcache_cache_rest;
-typedef struct mapcache_cache_s3 mapcache_cache_s3;
-typedef struct mapcache_cache_azure mapcache_cache_azure;
-typedef struct mapcache_cache_google mapcache_cache_google;
-#ifdef USE_TIFF
-typedef struct mapcache_cache_tiff mapcache_cache_tiff;
-#endif
 typedef struct mapcache_http mapcache_http;
 typedef struct mapcache_request mapcache_request;
 typedef struct mapcache_request_image mapcache_request_image;
 typedef struct mapcache_request_proxy mapcache_request_proxy;
 typedef struct mapcache_request_get_capabilities mapcache_request_get_capabilities;
-typedef struct mapcache_request_get_capabilities_demo mapcache_request_get_capabilities_demo;
-typedef struct mapcache_request_get_capabilities_wms mapcache_request_get_capabilities_wms;
-typedef struct mapcache_request_get_capabilities_wmts mapcache_request_get_capabilities_wmts;
 typedef struct mapcache_forwarding_rule mapcache_forwarding_rule;
-typedef struct mapcache_request_get_capabilities_tms mapcache_request_get_capabilities_tms;
-typedef struct mapcache_request_get_capabilities_kml mapcache_request_get_capabilities_kml;
 
 typedef struct mapcache_request_get_tile mapcache_request_get_tile;
 typedef struct mapcache_request_get_map mapcache_request_get_map;
 typedef struct mapcache_service mapcache_service;
-typedef struct mapcache_service_wms mapcache_service_wms;
-typedef struct mapcache_service_wmts mapcache_service_wmts;
-typedef struct mapcache_service_gmaps mapcache_service_gmaps;
-typedef struct mapcache_service_ve mapcache_service_ve;
-typedef struct mapcache_service_tms mapcache_service_tms;
-typedef struct mapcache_service_kml mapcache_service_kml;
-typedef struct mapcache_service_mapguide mapcache_service_mapguide;
-typedef struct mapcache_service_demo mapcache_service_demo;
 typedef struct mapcache_server_cfg mapcache_server_cfg;
 typedef struct mapcache_image mapcache_image;
 typedef struct mapcache_grid mapcache_grid;
@@ -158,17 +106,28 @@ typedef struct mapcache_grid_level mapcache_grid_level;
 typedef struct mapcache_grid_link mapcache_grid_link;
 typedef struct mapcache_context mapcache_context;
 typedef struct mapcache_dimension mapcache_dimension;
-typedef struct mapcache_dimension_time mapcache_dimension_time;
-typedef struct mapcache_timedimension mapcache_timedimension;
-typedef struct mapcache_dimension_intervals mapcache_dimension_intervals;
-typedef struct mapcache_dimension_values mapcache_dimension_values;
-typedef struct mapcache_dimension_sqlite mapcache_dimension_sqlite;
-typedef struct mapcache_dimension_regex mapcache_dimension_regex;
+typedef struct mapcache_requested_dimension mapcache_requested_dimension;
 typedef struct mapcache_extent mapcache_extent;
 typedef struct mapcache_extent_i mapcache_extent_i;
 typedef struct mapcache_connection_pool mapcache_connection_pool;
 typedef struct mapcache_locker mapcache_locker;
 
+typedef enum {
+  MAPCACHE_DIMENSION_VALUES,
+  MAPCACHE_DIMENSION_REGEX,
+  MAPCACHE_DIMENSION_INTERVALS,
+  MAPCACHE_DIMENSION_TIME,
+  MAPCACHE_DIMENSION_SQLITE
+} mapcache_dimension_type;
+
+typedef enum {
+  MAPCACHE_DIMENSION_ASSEMBLY_NONE,
+  MAPCACHE_DIMENSION_ASSEMBLY_STACK,
+  MAPCACHE_DIMENSION_ASSEMBLY_ANIMATE
+} mapcache_dimension_assembly_type;
+
+
+
 /** \defgroup utility Utility */
 /** @{ */
 
@@ -207,7 +166,7 @@ struct mapcache_context {
   void (*set_exception)(mapcache_context *ctx, char *key, char *message, ...);
 
   /**
-   * \brief query context to know if an error has occured
+   * \brief query context to know if an error has occurred
    * \memberof mapcache_context
    */
   int (*get_error)(mapcache_context * ctx);
@@ -247,7 +206,6 @@ struct mapcache_context {
   mapcache_context* (*clone)(mapcache_context *ctx);
   apr_pool_t *pool;
   mapcache_connection_pool *connection_pool;
-  void *threadlock;
   char *_contenttype;
   char *_errmsg;
   int _errcode;
@@ -307,7 +265,8 @@ typedef enum {
   MAPCACHE_SOURCE_WMS,
   MAPCACHE_SOURCE_MAPSERVER,
   MAPCACHE_SOURCE_DUMMY,
-  MAPCACHE_SOURCE_GDAL
+  MAPCACHE_SOURCE_GDAL,
+  MAPCACHE_SOURCE_FALLBACK
 } mapcache_source_type;
 
 /**\interface mapcache_source
@@ -318,6 +277,8 @@ struct mapcache_source {
   mapcache_extent data_extent; /**< extent in which this source can produce data */
   mapcache_source_type type;
   apr_table_t *metadata;
+  unsigned int retry_count;
+  double retry_delay;
 
   apr_array_header_t *info_formats;
   /**
@@ -325,11 +286,11 @@ struct mapcache_source {
    *
    * sets the mapcache_metatile::tile::data for the given tile
    */
-  void (*render_map)(mapcache_context *ctx, mapcache_map *map);
+  void (*_render_map)(mapcache_context *ctx, mapcache_source *psource, mapcache_map *map);
 
-  void (*query_info)(mapcache_context *ctx, mapcache_feature_info *fi);
+  void (*_query_info)(mapcache_context *ctx, mapcache_source *psource, mapcache_feature_info *fi);
 
-  void (*configuration_parse_xml)(mapcache_context *ctx, ezxml_t xml, mapcache_source * source);
+  void (*configuration_parse_xml)(mapcache_context *ctx, ezxml_t xml, mapcache_source * source, mapcache_cfg *config);
   void (*configuration_check)(mapcache_context *ctx, mapcache_cfg *cfg, mapcache_source * source);
 };
 
@@ -346,81 +307,23 @@ struct mapcache_http {
   /* TODO: authentication */
 };
 
-/**\class mapcache_source_wms
- * \brief WMS mapcache_source
- * \implements mapcache_source
- */
-struct mapcache_source_wms {
-  mapcache_source source;
-  apr_table_t *wms_default_params; /**< default WMS parameters (SERVICE,REQUEST,STYLES,VERSION) */
-  apr_table_t *getmap_params; /**< WMS parameters specified in configuration */
-  apr_table_t *getfeatureinfo_params; /**< WMS parameters specified in configuration */
-  mapcache_http *http;
-};
 
-#ifdef USE_MAPSERVER
-/**\class mapcache_source_mapserver
- * \brief WMS mapcache_source
- * \implements mapcache_source
- */
-typedef struct mapcache_source_mapserver mapcache_source_mapserver;
-struct mapcache_source_mapserver {
-  mapcache_source source;
-  char *mapfile;
-};
-#endif
-
-typedef struct mapcache_source_dummy mapcache_source_dummy;
-struct mapcache_source_dummy {
-  mapcache_source source;
-  char *mapfile;
-  void *mapobj;
-};
-#if 0
-#ifdef USE_GDAL
-/**\class mapcache_source_gdal
- * \brief GDAL mapcache_source
- * \implements mapcache_source
- */
-struct mapcache_source_gdal {
-  mapcache_source source;
-  char *datastr; /**< the gdal source string*/
-  apr_table_t *gdal_params; /**< GDAL parameters specified in configuration */
-  GDALDatasetH *poDataset;
-};
-#endif
-/** @} */
-#endif
 
 
 /** \defgroup cache Caches */
 
 /** @{ */
 typedef enum {
-  MAPCACHE_CACHE_DISK,
-  MAPCACHE_CACHE_REST
-#ifdef USE_MEMCACHE
+  MAPCACHE_CACHE_DISK
+  ,MAPCACHE_CACHE_REST
   ,MAPCACHE_CACHE_MEMCACHE
-#endif
-#ifdef USE_SQLITE
   ,MAPCACHE_CACHE_SQLITE
-#endif
-#ifdef USE_BDB
   ,MAPCACHE_CACHE_BDB
-#endif
-#ifdef USE_TC
   ,MAPCACHE_CACHE_TC
-#endif
-#ifdef USE_TIFF
   ,MAPCACHE_CACHE_TIFF
-#endif
   ,MAPCACHE_CACHE_COMPOSITE
-#ifdef USE_COUCHBASE
   ,MAPCACHE_CACHE_COUCHBASE
-#endif
-#ifdef USE_RIAK
   ,MAPCACHE_CACHE_RIAK
-#endif
 } mapcache_cache_type;
 
 /** \interface mapcache_cache
@@ -430,293 +333,77 @@ struct mapcache_cache {
   char *name; /**< key this cache is referenced by */
   mapcache_cache_type type;
   apr_table_t *metadata;
+  unsigned int retry_count;
+  double retry_delay;
+
 
   /**
    * get tile content from cache
-   * \returns MAPCACHE_SUCCESS if the data was correctly loaded from the disk
+   * \returns MAPCACHE_SUCCESS if the data was correctly loaded from the cache
    * \returns MAPCACHE_FAILURE if the file exists but contains no data
-   * \returns MAPCACHE_CACHE_MISS if the file does not exist on the disk
+   * \returns MAPCACHE_CACHE_MISS if the file does not exist on the cache
    * \memberof mapcache_cache
    */
-  int (*tile_get)(mapcache_context *ctx, mapcache_cache *cache, mapcache_tile * tile);
+  int (*_tile_get)(mapcache_context *ctx, mapcache_cache *cache, mapcache_tile * tile);
 
   /**
    * delete tile from cache
    *
    * \memberof mapcache_cache
    */
-  void (*tile_delete)(mapcache_context *ctx, mapcache_cache *cache, mapcache_tile * tile);
+  void (*_tile_delete)(mapcache_context *ctx, mapcache_cache *cache, mapcache_tile * tile);
 
-  int (*tile_exists)(mapcache_context *ctx, mapcache_cache *cache, mapcache_tile * tile);
+  int (*_tile_exists)(mapcache_context *ctx, mapcache_cache *cache, mapcache_tile * tile);
 
   /**
    * set tile content to cache
    * \memberof mapcache_cache
    */
-  void (*tile_set)(mapcache_context *ctx, mapcache_cache *cache, mapcache_tile * tile);
-  void (*tile_multi_set)(mapcache_context *ctx, mapcache_cache *cache, mapcache_tile *tiles, int ntiles);
+  void (*_tile_set)(mapcache_context *ctx, mapcache_cache *cache, mapcache_tile * tile);
+  void (*_tile_multi_set)(mapcache_context *ctx, mapcache_cache *cache, mapcache_tile *tiles, int ntiles);
 
   void (*configuration_parse_xml)(mapcache_context *ctx, ezxml_t xml, mapcache_cache * cache, mapcache_cfg *config);
   void (*configuration_post_config)(mapcache_context *ctx, mapcache_cache * cache, mapcache_cfg *config);
 };
 
-/**\class mapcache_cache_disk
- * \brief a mapcache_cache on a filesytem
- * \implements mapcache_cache
- */
-struct mapcache_cache_disk {
-  mapcache_cache cache;
-  char *base_directory;
-  char *filename_template;
-  int symlink_blank;
-  int creation_retry;
-
-  /**
-   * Set filename for a given tile
-   * \memberof mapcache_cache_disk
-   */
-  void (*tile_key)(mapcache_context *ctx, mapcache_cache_disk *cache, mapcache_tile *tile, char **path);
-};
-
-typedef struct mapcache_cache_composite_cache_link mapcache_cache_composite_cache_link;
-struct mapcache_cache_composite_cache_link {
-  mapcache_cache *cache;
-  int minzoom;
-  int maxzoom;
-  apr_array_header_t *grids;
-  apr_array_header_t *dimensions; //TODO
-};
-
-struct mapcache_cache_composite {
-  mapcache_cache cache;
-  apr_array_header_t *cache_links;
-};
-
-struct mapcache_cache_fallback {
-  mapcache_cache cache;
-  apr_array_header_t *caches;
-};
-
-struct mapcache_cache_multitier {
-  mapcache_cache cache;
-  apr_array_header_t *caches;
-};
+MS_DLL_EXPORT int mapcache_cache_tile_get(mapcache_context *ctx, mapcache_cache *cache, mapcache_tile *tile);
+void mapcache_cache_tile_delete(mapcache_context *ctx, mapcache_cache *cache, mapcache_tile *tile);
+MS_DLL_EXPORT int mapcache_cache_tile_exists(mapcache_context *ctx, mapcache_cache *cache, mapcache_tile *tile);
+MS_DLL_EXPORT void mapcache_cache_tile_set(mapcache_context *ctx, mapcache_cache *cache, mapcache_tile *tile);
+void mapcache_cache_tile_multi_set(mapcache_context *ctx, mapcache_cache *cache, mapcache_tile *tiles, int ntiles);
 
 
-typedef enum {
-  MAPCACHE_REST_METHOD_GET,
-  MAPCACHE_REST_METHOD_HEAD,
-  MAPCACHE_REST_METHOD_PUT,
-  MAPCACHE_REST_METHOD_POST,
-  MAPCACHE_REST_METHOD_DELETE
-} mapcache_rest_method;
-
-typedef enum {
-  MAPCACHE_REST_PROVIDER_NONE,
-  MAPCACHE_REST_PROVIDER_S3,
-  MAPCACHE_REST_PROVIDER_AZURE,
-  MAPCACHE_REST_PROVIDER_GOOGLE
-} mapcache_rest_provider;
-
-void sha256(const unsigned char *message, unsigned int len, unsigned char *digest);
-void hmac_sha256(const unsigned char *message, unsigned int message_len,
-          const unsigned char *key, unsigned int key_size,
-          unsigned char *mac, unsigned mac_size);
-void hmac_sha1(const char *message, unsigned int message_len,
-          const unsigned char *key, unsigned int key_size,
-          void *mac);
-void sha_hex_encode(unsigned char *sha, unsigned int sha_size);
-char *base64_encode(apr_pool_t *pool, const unsigned char *data, size_t input_length);
-
-typedef struct mapcache_rest_operation mapcache_rest_operation;
-struct mapcache_rest_operation {
-  apr_table_t *headers;
-  mapcache_rest_method method;
-  char *tile_url;
-  void (*add_headers)(mapcache_context *ctx, mapcache_cache_rest *pcache, mapcache_tile *tile, char *url, apr_table_t *headers);
-};
-
-typedef struct mapcache_rest_configuration mapcache_rest_configuration;
-struct mapcache_rest_configuration {
-  apr_table_t *common_headers;
-  char *tile_url;
-  mapcache_rest_operation has_tile;
-  mapcache_rest_operation get_tile;
-  mapcache_rest_operation set_tile;
-  mapcache_rest_operation multi_set_tile;
-  mapcache_rest_operation delete_tile;
-  void (*add_headers)(mapcache_context *ctx, mapcache_cache_rest *pcache, mapcache_tile *tile, char *url, apr_table_t *headers);
-};
-
-/**\class mapcache_cache_rest
- * \brief a mapcache_cache on a 3rd party HTTP Rest API
- * \implements mapcache_cache
- */
-struct mapcache_cache_rest {
-  mapcache_cache cache;
-  mapcache_rest_configuration rest;
-  int use_redirects;
-  int retry_count;
-  mapcache_rest_provider provider;
-};
-
-struct mapcache_cache_s3 {
-  mapcache_cache_rest cache;
-  char *id;
-  char *secret;
-  char *region;
-};
-
-struct mapcache_cache_azure {
-  mapcache_cache_rest cache;
-  char *id;
-  char *secret;
-  char *container;
-};
-
-struct mapcache_cache_google {
-  mapcache_cache_rest cache;
-  char *access;
-  char *secret;
-};
-
-#ifdef USE_TIFF
-struct mapcache_cache_tiff {
-  mapcache_cache cache;
-  char *filename_template;
-  char *x_fmt,*y_fmt,*z_fmt,*inv_x_fmt,*inv_y_fmt,*div_x_fmt,*div_y_fmt,*inv_div_x_fmt,*inv_div_y_fmt;
-  int count_x;
-  int count_y;
-  mapcache_image_format_jpeg *format;
-  mapcache_locker *locker;
-};
-#endif
-
-#ifdef USE_SQLITE
-/**\class mapcache_cache_sqlite
- * \brief a mapcache_cache on a filesytem
- * \implements mapcache_cache
- */
-typedef struct mapcache_cache_sqlite mapcache_cache_sqlite;
-typedef struct mapcache_cache_sqlite_stmt mapcache_cache_sqlite_stmt;
-
-struct mapcache_cache_sqlite_stmt {
-  char *sql;
-};
-
-struct sqlite_conn;
-
-struct mapcache_cache_sqlite {
-  mapcache_cache cache;
-  char *dbfile;
-  mapcache_cache_sqlite_stmt create_stmt;
-  mapcache_cache_sqlite_stmt exists_stmt;
-  mapcache_cache_sqlite_stmt get_stmt;
-  mapcache_cache_sqlite_stmt set_stmt;
-  mapcache_cache_sqlite_stmt delete_stmt;
-  apr_table_t *pragmas;
-  void (*bind_stmt)(mapcache_context *ctx, void *stmt, mapcache_cache_sqlite *cache, mapcache_tile *tile);
-  int n_prepared_statements;
-  int detect_blank;
-  char *x_fmt,*y_fmt,*z_fmt,*inv_x_fmt,*inv_y_fmt,*div_x_fmt,*div_y_fmt,*inv_div_x_fmt,*inv_div_y_fmt;
-  int count_x, count_y;
-};
 
 /**
  * \memberof mapcache_cache_sqlite
  */
 mapcache_cache* mapcache_cache_sqlite_create(mapcache_context *ctx);
 mapcache_cache* mapcache_cache_mbtiles_create(mapcache_context *ctx);
-#endif
 
-#ifdef USE_BDB
-typedef struct mapcache_cache_bdb mapcache_cache_bdb;
-struct mapcache_cache_bdb {
-  mapcache_cache cache;
-  char *basedir;
-  char *key_template;
-};
-mapcache_cache *mapcache_cache_bdb_create(mapcache_context *ctx);
-#endif
-
-#ifdef USE_TC
-typedef struct mapcache_cache_tc mapcache_cache_tc;
-struct mapcache_cache_tc {
-  mapcache_cache cache;
-  char *basedir;
-  char *key_template;
-  mapcache_context *ctx;
-};
-mapcache_cache *mapcache_cache_tc_create(mapcache_context *ctx);
-#endif
-
-#ifdef USE_MEMCACHE
-typedef struct mapcache_cache_memcache mapcache_cache_memcache;
-/**\class mapcache_cache_memcache
- * \brief a mapcache_cache on memcached servers
- * \implements mapcache_cache
+/**
+ * \memberof mapcache_cache_bdb
  */
-
-struct mapcache_cache_memcache_server {
-    char* host;
-    int port;
-};
-
-struct mapcache_cache_memcache {
-  mapcache_cache cache;
-  int nservers;
-  struct mapcache_cache_memcache_server *servers;
-  int detect_blank;
-};
+mapcache_cache *mapcache_cache_bdb_create(mapcache_context *ctx);
 
 /**
  * \memberof mapcache_cache_memcache
  */
 mapcache_cache* mapcache_cache_memcache_create(mapcache_context *ctx);
-#endif
-
-#ifdef USE_COUCHBASE
-typedef struct mapcache_cache_couchbase mapcache_cache_couchbase;
-
-/**\class mapcache_cache_couchbase
- * \brief a mapcache_cache on couchbase servers
- * \implements mapcache_cache
- */
-struct mapcache_cache_couchbase {
-   mapcache_cache cache;
-//   apr_reslist_t *connection_pool;
-   char *host;
-   char *username;
-   char *password;
-   char *bucket;
-   mapcache_context *ctx;
-};
 
 /**
  * \memberof mapcache_cache_couchbase
  */
 mapcache_cache* mapcache_cache_couchbase_create(mapcache_context *ctx);
-#endif
 
-#ifdef USE_RIAK
-typedef struct mapcache_cache_riak mapcache_cache_riak;
-
-/**\class mapcache_cache_riak
- * \brief a mapcache_cache for riak servers
- * \implements mapcache_cache
+/**
+ * \memberof mapcache_cache_tc
  */
-struct mapcache_cache_riak {
-   mapcache_cache cache;
-   char *host;
-   int port;
-   RIACK_STRING bucket;
-};
+mapcache_cache* mapcache_cache_tc_create(mapcache_context *ctx);
 
 /**
  * \memberof mapcache_cache_riak
  */
 mapcache_cache* mapcache_cache_riak_create(mapcache_context *ctx);
-#endif
 
 /** @} */
 
@@ -783,7 +470,7 @@ struct mapcache_http_response {
 struct mapcache_map {
   mapcache_tileset *tileset;
   mapcache_grid_link *grid_link;
-  apr_table_t *dimensions;
+  apr_array_header_t *dimensions;
   mapcache_buffer *encoded_data;
   mapcache_image *raw_image;
   int nodata; /**< \sa mapcache_tile::nodata */
@@ -827,36 +514,7 @@ struct mapcache_request_get_capabilities {
   char *mime_type;
 };
 
-struct mapcache_request_get_capabilities_tms {
-  mapcache_request_get_capabilities request;
-  mapcache_tileset *tileset;
-  mapcache_grid_link *grid_link;
-  char *version;
-};
 
-struct mapcache_request_get_capabilities_kml {
-  mapcache_request_get_capabilities request;
-  mapcache_tile *tile;
-  mapcache_tileset *tileset;
-  mapcache_grid_link *grid;
-};
-
-struct mapcache_request_get_capabilities_wms {
-  mapcache_request_get_capabilities request;
-};
-
-struct mapcache_request_get_capabilities_wmts {
-  mapcache_request_get_capabilities request;
-};
-
-/**
- * the capabilities request for a specific service, to be able to create
- * demo pages specific to a given service
- */
-struct mapcache_request_get_capabilities_demo {
-  mapcache_request_get_capabilities request;
-  mapcache_service *service;
-};
 
 struct mapcache_forwarding_rule {
   char *name;
@@ -932,67 +590,9 @@ struct mapcache_service {
                        char **err_body, apr_table_t *headers);
 };
 
-/**\class mapcache_service_wms
- * \brief an OGC WMS service
- * \implements mapcache_service
- */
-struct mapcache_service_wms {
-  mapcache_service service;
-  int maxsize;
-  apr_array_header_t *forwarding_rules;
-  mapcache_getmap_strategy getmap_strategy;
-  mapcache_resample_mode resample_mode;
-  mapcache_image_format *getmap_format;
-  int allow_format_override; /* can the client specify which image format should be returned */
-};
 
-/**\class mapcache_service_kml
- * \brief a KML superoverlay service
- * \implements mapcache_service
- */
-struct mapcache_service_kml {
-  mapcache_service service;
-};
 
-/**\class mapcache_service_tms
- * \brief a TMS service
- * \implements mapcache_service
- */
-struct mapcache_service_tms {
-  mapcache_service service;
-  int reverse_y;
-};
 
-struct mapcache_service_mapguide {
-  mapcache_service service;
-  int rows_per_folder;
-  int cols_per_folder;
-};
-
-/**\class mapcache_service_wmts
- * \brief a WMTS service
- * \implements mapcache_service
- */
-struct mapcache_service_wmts {
-  mapcache_service service;
-};
-
-/**\class mapcache_service_demo
- * \brief a demo service
- * \implements mapcache_service
- */
-struct mapcache_service_demo {
-  mapcache_service service;
-
-};
-
-/**\class mapcache_service_ve
- * \brief a virtualearth service
- * \implements mapcache_service
- */
-struct mapcache_service_ve {
-  mapcache_service service;
-};
 
 /**
  * \brief create and initialize a mapcache_service_wms
@@ -1050,6 +650,10 @@ MS_DLL_EXPORT void mapcache_service_dispatch_request(mapcache_context *ctx,
                                        char *pathinfo,
                                        apr_table_t *params,
                                        mapcache_cfg *config);
+/**
+ * \brief return the number of enabled/configured services
+ */
+MS_DLL_EXPORT int mapcache_config_services_enabled(mapcache_context *ctx, mapcache_cfg *config);
 
 
 /** @} */
@@ -1140,7 +744,7 @@ int mapcache_image_blank_color(mapcache_image* image);
 /**
  * \brief check if image has some non opaque pixels
  */
-int mapcache_image_has_alpha(mapcache_image *img);
+int mapcache_image_has_alpha(mapcache_image *img, unsigned int cutoff);
 
 void mapcache_image_fill(mapcache_context *ctx, mapcache_image *image, const unsigned char *fill_color);
 
@@ -1182,8 +786,8 @@ typedef enum {
 
 struct mapcache_locker{
   mapcache_lock_result (*aquire_lock)(mapcache_context *ctx, mapcache_locker *self, char *resource, void **lock);
-  mapcache_lock_result (*ping_lock)(mapcache_context *ctx, mapcache_locker *self, char *resource, void *lock);
-  void (*release_lock)(mapcache_context *ctx, mapcache_locker *self, char *resource, void *lock);
+  mapcache_lock_result (*ping_lock)(mapcache_context *ctx, mapcache_locker *self, void *lock);
+  void (*release_lock)(mapcache_context *ctx, mapcache_locker *self, void *lock);
 
   void (*parse_xml)(mapcache_context *ctx, mapcache_locker *self, ezxml_t node);
   mapcache_lock_mode type;
@@ -1191,43 +795,14 @@ struct mapcache_locker{
   double retry_interval; /* time to wait before checking again on a lock, in seconds */
 };
 
-typedef struct {
-  mapcache_locker locker;
-
-  /**
-   * directory where lock files will be placed.
-   * Must be readable and writable by the apache user.
-   * Must be placed on a network mounted shared directory if multiple mapcache instances
-   * need to be synchronized
-   */
-  const char *dir;
-} mapcache_locker_disk;
-
 mapcache_locker* mapcache_locker_disk_create(mapcache_context *ctx);
 
-#ifdef USE_MEMCACHE
-typedef struct {
-  char *host;
-  int port;
-} mapcache_locker_memcache_server;
-
-typedef struct {
-  mapcache_locker locker;
-  int nservers;
-  mapcache_locker_memcache_server *servers;
-} mapcache_locker_memcache;
-
 mapcache_locker* mapcache_locker_memcache_create(mapcache_context *ctx);
-#endif
-
-typedef struct {
-    mapcache_locker locker;
-    apr_array_header_t *lockers;
-} mapcache_locker_fallback;
 
 mapcache_locker* mapcache_locker_fallback_create(mapcache_context *ctx);
 
 void mapcache_config_parse_locker(mapcache_context *ctx, ezxml_t node, mapcache_locker **locker);
+void mapcache_config_parse_locker_old(mapcache_context *ctx, ezxml_t doc, mapcache_cfg *config);
 
 /**
  * a configuration that will be served
@@ -1329,22 +904,32 @@ void mapcache_configuration_add_cache(mapcache_cfg *config, mapcache_cache *cach
  */
 void mapcache_source_init(mapcache_context *ctx, mapcache_source *source);
 
+/**
+ * \memberof mapcache_source
+ */
+void mapcache_source_render_map(mapcache_context *ctx, mapcache_source *source, mapcache_map *map);
+void mapcache_source_query_info(mapcache_context *ctx, mapcache_source *source,
+    mapcache_feature_info *fi);
+
 /**
  * \memberof mapcache_source_gdal
  */
 mapcache_source* mapcache_source_gdal_create(mapcache_context *ctx);
 
+/**
+ * \memberof mapcache_source_fallback
+ */
+mapcache_source* mapcache_source_fallback_create(mapcache_context *ctx);
+
 /**
  * \memberof mapcache_source_wms
  */
 mapcache_source* mapcache_source_wms_create(mapcache_context *ctx);
 
-#ifdef USE_MAPSERVER
 /**
  * \memberof mapcache_source_wms
  */
 mapcache_source* mapcache_source_mapserver_create(mapcache_context *ctx);
-#endif
 
 mapcache_source* mapcache_source_dummy_create(mapcache_context *ctx);
 
@@ -1361,12 +946,10 @@ mapcache_cache* mapcache_cache_s3_create(mapcache_context *ctx);
 mapcache_cache* mapcache_cache_azure_create(mapcache_context *ctx);
 mapcache_cache* mapcache_cache_google_create(mapcache_context *ctx);
 
-#ifdef USE_TIFF
 /**
  * \memberof mapcache_cache_tiff
  */
 mapcache_cache* mapcache_cache_tiff_create(mapcache_context *ctx);
-#endif
 
 mapcache_cache* mapcache_cache_composite_create(mapcache_context *ctx);
 mapcache_cache* mapcache_cache_fallback_create(mapcache_context *ctx);
@@ -1400,7 +983,8 @@ struct mapcache_tile {
   apr_time_t mtime; /**< last modification time */
   int expires; /**< time in seconds after which the tile should be rechecked for validity */
 
-  apr_table_t *dimensions;
+  apr_array_header_t *dimensions;
+
   /**
    * flag stating the tile is empty (i.e. fully transparent).
    * if set, this indicates that there was no error per se, but that there was
@@ -1531,6 +1115,7 @@ struct mapcache_tileset {
   int auto_expire;
 
   int read_only;
+  int subdimension_read_only;
 
   /**
    * the cache in which the tiles should be stored
@@ -1552,7 +1137,9 @@ struct mapcache_tileset {
    */
   apr_array_header_t *dimensions;
 
-  mapcache_timedimension *timedimension;
+  int store_dimension_assemblies;/**< should multiple sub-dimensions be assembled dynamically (per-request) or should they be cached once assembled */
+
+  mapcache_dimension_assembly_type dimension_assembly_type;
 
   /**
    * image to be used as a watermark
@@ -1599,6 +1186,9 @@ int mapcache_grid_get_cell(mapcache_context *ctx, mapcache_grid *grid, mapcache_
  * @return
  */
 void mapcache_tileset_tile_validate(mapcache_context *ctx, mapcache_tile *tile);
+void mapcache_tileset_tile_validate_z(mapcache_context *ctx, mapcache_tile *tile);
+void mapcache_tileset_tile_validate_x(mapcache_context *ctx, mapcache_tile *tile);
+void mapcache_tileset_tile_validate_y(mapcache_context *ctx, mapcache_tile *tile);
 
 /**
  * compute level for a given resolution
@@ -1614,6 +1204,7 @@ void mapcache_tileset_get_level(mapcache_context *ctx, mapcache_tileset *tileset
 
 mapcache_grid_link* mapcache_grid_get_closest_wms_level(mapcache_context *ctx, mapcache_grid_link *grid, double resolution, int *level);
 MS_DLL_EXPORT void mapcache_tileset_tile_get(mapcache_context *ctx, mapcache_tile *tile);
+MS_DLL_EXPORT void mapcache_tileset_tile_set_get_with_subdimensions(mapcache_context *ctx, mapcache_tile *tile);
 
 /**
  * \brief delete tile from cache
@@ -1664,7 +1255,7 @@ void mapcache_tileset_add_watermark(mapcache_context *ctx, mapcache_tileset *til
 
 
 MS_DLL_EXPORT int mapcache_lock_or_wait_for_resource(mapcache_context *ctx, mapcache_locker *locker, char *resource, void **lock);
-MS_DLL_EXPORT void mapcache_unlock_resource(mapcache_context *ctx, mapcache_locker *locker, char *resource, void *lock);
+MS_DLL_EXPORT void mapcache_unlock_resource(mapcache_context *ctx, mapcache_locker *locker, void *lock);
 
 MS_DLL_EXPORT mapcache_metatile* mapcache_tileset_metatile_get(mapcache_context *ctx, mapcache_tile *tile);
 MS_DLL_EXPORT void mapcache_tileset_render_metatile(mapcache_context *ctx, mapcache_metatile *mt);
@@ -1692,8 +1283,11 @@ mapcache_grid* mapcache_grid_create(apr_pool_t *pool);
 const char* mapcache_grid_get_crs(mapcache_context *ctx, mapcache_grid *grid);
 const char* mapcache_grid_get_srs(mapcache_context *ctx, mapcache_grid *grid);
 
-MS_DLL_EXPORT void mapcache_grid_get_extent(mapcache_context *ctx, mapcache_grid *grid,
+MS_DLL_EXPORT void mapcache_grid_get_tile_extent(mapcache_context *ctx, mapcache_grid *grid,
                               int x, int y, int z, mapcache_extent *bbox);
+
+MS_DLL_EXPORT void mapcache_grid_get_metatile_extent(mapcache_context *ctx, mapcache_tile *tile, mapcache_extent *bbox);
+
 /**
  * \brief compute x y value for given lon/lat (dx/dy) and given zoomlevel
  * @param ctx
@@ -1733,6 +1327,9 @@ MS_DLL_EXPORT int mapcache_util_extract_double_list(mapcache_context *ctx, const
                                       int *numbers_count);
 char *mapcache_util_str_replace(apr_pool_t *pool, const char *string, const char *substr,
                                 const char *replacement );
+void mapcache_util_quadkey_decode(mapcache_context *ctx, const char *quadkey, int *x, int *y, int *z);
+
+char* mapcache_util_quadkey_encode(mapcache_context *ctx, int x, int y, int z);
 
 /**
  * \brief replace dangerous characters in string
@@ -1771,6 +1368,15 @@ typedef enum {
   MAPCACHE_PHOTOMETRIC_YCBCR
 } mapcache_photometric;
 
+/**
+ * optimization settings (mostly) for jpeg
+ */
+typedef enum {
+  MAPCACHE_OPTIMIZE_NO,
+  MAPCACHE_OPTIMIZE_YES,
+  MAPCACHE_OPTIMIZE_ARITHMETIC
+} mapcache_optimization;
+
 /**\interface mapcache_image_format
  * \brief an image format
  * \sa mapcache_image_format_jpeg
@@ -1809,13 +1415,14 @@ struct mapcache_image_format_mixed {
   mapcache_image_format format;
   mapcache_image_format *transparent;
   mapcache_image_format *opaque;
+  unsigned int alpha_cutoff; /* default 255. pixels with alpha >= alpha_cutoff will be considered fully opaque */
 };
 
 mapcache_buffer* mapcache_empty_png_decode(mapcache_context *ctx, int width, int height, const unsigned char *hex_color, int *is_empty);
 
 
 mapcache_image_format* mapcache_imageio_create_mixed_format(apr_pool_t *pool,
-    char *name, mapcache_image_format *transparent, mapcache_image_format *opaque);
+    char *name, mapcache_image_format *transparent, mapcache_image_format *opaque, unsigned int alpha_cutoff);
 
 /**\class mapcache_image_format_png_q
  * \brief Quantized PNG format
@@ -1878,10 +1485,11 @@ struct mapcache_image_format_jpeg {
   mapcache_image_format format;
   int quality; /**< JPEG quality, 1-100 */
   mapcache_photometric photometric;
+  mapcache_optimization optimize;
 };
 
 mapcache_image_format* mapcache_imageio_create_jpeg_format(apr_pool_t *pool, char *name, int quality,
-    mapcache_photometric photometric);
+    mapcache_photometric photometric, mapcache_optimization optimize);
 
 /**
  * @param r
@@ -1930,13 +1538,19 @@ typedef struct {
   double resolution;
 } mapcache_interval;
 
-typedef enum {
-  MAPCACHE_DIMENSION_VALUES,
-  MAPCACHE_DIMENSION_REGEX,
-  MAPCACHE_DIMENSION_INTERVALS,
-  MAPCACHE_DIMENSION_TIME,
-  MAPCACHE_DIMENSION_SQLITE
-} mapcache_dimension_type;
+struct mapcache_requested_dimension {
+  mapcache_dimension *dimension;
+  char *requested_value;
+  char *cached_value;
+};
+
+void mapcache_tile_set_cached_dimension(mapcache_context *ctx, mapcache_tile *tile, const char *name, const char *value);
+void mapcache_map_set_cached_dimension(mapcache_context *ctx, mapcache_map *map, const char *name, const char *value);
+void mapcache_tile_set_requested_dimension(mapcache_context *ctx, mapcache_tile *tile, const char *name, const char *value);
+void mapcache_map_set_requested_dimension(mapcache_context *ctx, mapcache_map *map, const char *name, const char *value);
+void mapcache_set_requested_dimension(mapcache_context *ctx, apr_array_header_t *dimensions, const char *name, const char *value);
+void mapcache_set_cached_dimension(mapcache_context *ctx, apr_array_header_t *dimensions, const char *name, const char *value);
+MS_DLL_EXPORT apr_array_header_t *mapcache_requested_dimensions_clone(apr_pool_t *pool, apr_array_header_t *src);
 
 struct mapcache_dimension {
   mapcache_dimension_type type;
@@ -1944,24 +1558,24 @@ struct mapcache_dimension {
   char *unit;
   apr_table_t *metadata;
   char *default_value;
-  int skip_validation;
 
   /**
-   * \brief validate the given value
-   *
-   * \param value is updated in case the given value is correct but has to be represented otherwise,
-   * e.g. to round off a value
-   * \returns MAPCACHE_SUCCESS if the given value is correct for the current dimension
-   * \returns MAPCACHE_FAILURE if not
+   * \brief return the list of dimension values that match the requested entry
    */
-  int (*validate)(mapcache_context *context, mapcache_dimension *dimension, char **value);
+  apr_array_header_t* (*get_entries_for_value)(mapcache_context *ctx, mapcache_dimension *dimension, const char *value,
+                       mapcache_tileset *tileset, mapcache_extent *extent, mapcache_grid *grid);
 
   /**
-   * \brief returns a list of values that are authorized for this dimension
-   *
-   * \returns a list of character strings that will be included in the capabilities <dimension> element
+   * \brief return all possible values
    */
-  apr_array_header_t*  (*print_ogc_formatted_values)(mapcache_context *context, mapcache_dimension *dimension);
+  apr_array_header_t* (*get_all_entries)(mapcache_context *ctx, mapcache_dimension *dimension,
+                       mapcache_tileset *tileset, mapcache_extent *extent, mapcache_grid *grid);
+
+  /**
+   * \brief return all possible values formatted in a way compatible with OGC capabilities <dimension> element
+   */
+  apr_array_header_t* (*get_all_ogc_formatted_entries)(mapcache_context *ctx, mapcache_dimension *dimension,
+                       mapcache_tileset *tileset, mapcache_extent *extent, mapcache_grid *grid);
 
   /**
    * \brief parse the value given in the configuration
@@ -1969,79 +1583,10 @@ struct mapcache_dimension {
   void (*configuration_parse_xml)(mapcache_context *context, mapcache_dimension *dim, ezxml_t node);
 };
 
-struct mapcache_dimension_values {
-  mapcache_dimension dimension;
-  int nvalues;
-  char **values;
-  int case_sensitive;
-};
-
-struct mapcache_dimension_sqlite {
-  mapcache_dimension dimension;
-  char *dbfile;
-  char *validate_query;
-  char *list_query;
-};
-
-struct mapcache_dimension_regex {
-  mapcache_dimension dimension;
-  char *regex_string;
-#ifdef USE_PCRE
-  pcre *pcregex;
-#else
-  regex_t *regex;
-#endif
-};
-
-struct mapcache_dimension_intervals {
-  mapcache_dimension dimension;
-  int nintervals;
-  mapcache_interval *intervals;
-};
-
-struct mapcache_dimension_time {
-  mapcache_dimension dimension;
-  int nintervals;
-  mapcache_interval *intervals;
-};
-
-mapcache_dimension* mapcache_dimension_values_create(apr_pool_t *pool);
-mapcache_dimension* mapcache_dimension_sqlite_create(apr_pool_t *pool);
-mapcache_dimension* mapcache_dimension_regex_create(apr_pool_t *pool);
-mapcache_dimension* mapcache_dimension_intervals_create(apr_pool_t *pool);
-mapcache_dimension* mapcache_dimension_time_create(apr_pool_t *pool);
-
-typedef enum {
-  MAPCACHE_TIMEDIMENSION_ASSEMBLY_STACK,
-  MAPCACHE_TIMEDIMENSION_ASSEMBLY_ANIMATE
-} mapcache_timedimension_assembly_type;
-
-typedef enum {
-  MAPCACHE_TIMEDIMENSION_SOURCE_SQLITE
-} mapcache_timedimension_source_type;
-
-MS_DLL_EXPORT apr_array_header_t* mapcache_timedimension_get_entries_for_value(mapcache_context *ctx, mapcache_timedimension *timedimesnion,
-        mapcache_tileset *tileset, mapcache_grid *grid, mapcache_extent *extent, const char *value);
-
-struct mapcache_timedimension {
-  mapcache_timedimension_assembly_type assembly_type;
-  void (*configuration_parse_xml)(mapcache_context *context, mapcache_timedimension *dim, ezxml_t node);
-  apr_array_header_t* (*get_entries_for_interval)(mapcache_context *ctx, mapcache_timedimension *dim, mapcache_tileset *tileset,
-        mapcache_grid *grid, mapcache_extent *extent, time_t start, time_t end);
-  apr_array_header_t* (*get_all_entries)(mapcache_context *ctx, mapcache_timedimension *dim, mapcache_tileset *tileset);
-  char *default_value;
-  char *key; /* TIME, hardcoded */
-};
-
-#ifdef USE_SQLITE
-typedef struct mapcache_timedimension_sqlite mapcache_timedimension_sqlite;
-struct mapcache_timedimension_sqlite {
-  mapcache_timedimension timedimension;
-  char *dbfile;
-  char *query;
-};
-mapcache_timedimension* mapcache_timedimension_sqlite_create(apr_pool_t *pool);
-#endif
+mapcache_dimension* mapcache_dimension_values_create(mapcache_context *ctx, apr_pool_t *pool);
+mapcache_dimension* mapcache_dimension_sqlite_create(mapcache_context *ctx, apr_pool_t *pool);
+mapcache_dimension* mapcache_dimension_regex_create(mapcache_context *ctx, apr_pool_t *pool);
+mapcache_dimension* mapcache_dimension_time_create(mapcache_context *ctx, apr_pool_t *pool);
 
 int mapcache_is_axis_inverted(const char *srs);
 
